import unittest
from xaal.lib import tools
from xaal.lib import bindings


class TestTools(unittest.TestCase):


    def test_random(self):
        value = tools.get_random_uuid()
        self.assertIsInstance(value, bindings.UUID)

        value = tools.get_random_base_uuid()
        self.assertIsInstance(value, bindings.UUID)

    def test_get_uuid(self):
        uuid = bindings.UUID.random()
        addr = tools.get_uuid(uuid)
        self.assertEqual(addr, uuid)

        addr = tools.get_uuid("12345678-1234-1234-1234-123456789012")
        self.assertEqual(addr, bindings.UUID("12345678-1234-1234-1234-123456789012"))
        self.assertEqual(None, tools.get_uuid(12))

    def test_str_to_uuid(self):
        addr = tools.get_uuid("12345678-1234-1234-1234-123456789012")
        self.assertEqual(tools.str_to_uuid("12345678-1234-1234-1234-123456789012"), addr)
        self.assertEqual(tools.str_to_uuid("0"),None)

    def test_bytes_to_uuid(self):
        addr = bindings.UUID("12345678-1234-1234-1234-123456789012")
        value = tools.bytes_to_uuid(addr.bytes)
        self.assertEqual(value, addr)
        self.assertEqual(tools.bytes_to_uuid(b'0000'), None)

    def test_pass2key(self):
        key = "test"
        value = tools.pass2key(key)
        self.assertEqual(value, b'q\xc1\xc8]\x8ey\xd5\x8cC\xf5CF\xe67\xab\x1c\xd0\xfa\x82\r\x0br6a\xca]\\\xb2@6\xaf\xcf')

    def test_is_valid_dev_type(self):
        self.assertTrue(tools.is_valid_dev_type("light.basic"))
        self.assertFalse(tools.is_valid_dev_type("light"))
        self.assertFalse(tools.is_valid_dev_type(42))

    def test_reduce_addr(self):
        addr = bindings.UUID("12345678-1234-1234-1234-123456789012")
        value = tools.reduce_addr(addr)
        self.assertEqual(value, '12345..89012')


if __name__ == '__main__':
    unittest.main()


from io import BytesIO
from typing import Any
import cbor2
from cbor2 import CBORTag, CBORDecoder

from . import bindings


def tag_hook(decoder, tag, shareable_index=None):
    if tag.tag == 37:
        return bindings.UUID(bytes=tag.value)
    if tag.tag == 32:
        return bindings.URL(tag.value)
    return tag


def default_encoder(encoder, value):
    if isinstance(value, bindings.UUID):
        encoder.encode(CBORTag(37, value.bytes))

    if isinstance(value, bindings.URL):
        encoder.encode(CBORTag(32, value.bytes))


def dumps(obj, **kwargs):
    return cbor2.dumps(obj, default=default_encoder, **kwargs)


def loads(payload, **kwargs):
    # return cbor2.loads(payload,tag_hook=tag_hook,**kwargs)
    return _loads(payload, tag_hook=tag_hook, **kwargs)


def _loads(s, **kwargs):
    with BytesIO(s) as fp:
        # use a custom decoder in pure Py. C version doesn't support
        # subclassing, and enforce usage of builtin semantic tags
        dec = CustomDecoder(fp, **kwargs)
        return dec.decode()


class CustomDecoder(CBORDecoder):
    """Custom CBOR decoder, this version simply drop the builtin semantic tags"""

    def decode_semantic(self, subtype: int) -> Any:
        tagnum = self._decode_length(subtype)

        tag = CBORTag(tagnum, None)
        self.set_shareable(tag)
        tag.value = self._decode(unshared=True)
        if self._tag_hook:
            tag = self._tag_hook(self, tag)


def cleanup(obj):
    """
    recursive walk a object to search for un-wanted CBOR tags.
    Transform this tag in string format, this can be UUID, URL..
    Should be Ok, with list, dicts..
    Warning: This operate in-place changes.
    Warning: This won't work for tags in dict keys.
    """
    if isinstance(obj, list):
        for i in range(0, len(obj)):
            obj[i] = cleanup(obj[i])
        return obj

    if isinstance(obj, dict):
        for k in obj.keys():
            obj.update({k: cleanup(obj[k])})
        return obj

    if type(obj) in bindings.classes:
        return str(obj)
    else:
        return obj



import collections
import logging
import time
import typing
from enum import Enum
from typing import Optional

from .config import config
from . import core
from .exceptions import CallbackError, MessageParserError, XAALError
from .network import NetworkConnector

if typing.TYPE_CHECKING:
    from .devices import Device
    from .messages import Message


logger = logging.getLogger(__name__)


class EngineState(Enum):
    started = 1
    running = 2
    halted = 3


class Engine(core.EngineMixin):
    __slots__ = ['__last_timer', '__txFifo', 'state', 'network']

    def __init__(
        self, address: str = config.address, port: int = config.port, hops: int = config.hops, key: bytes = config.key
    ):
        core.EngineMixin.__init__(self, address, port, hops, key)

        self.__last_timer = 0  # last timer check
        self.__txFifo = collections.deque()  # tx msg fifo
        # message receive workflow
        self.subscribe(self.handle_request)
        # ready to go
        self.state = EngineState.halted
        # start network
        self.network = NetworkConnector(address, port, hops)

    #####################################################
    # xAAL messages Tx handling
    #####################################################
    # Fifo for msg to send
    def queue_msg(self, msg: bytes):
        """queue an encoded / cyphered message"""
        self.__txFifo.append(msg)

    def send_msg(self, msg: bytes):
        """Send an encoded message to the bus, use queue_msg instead"""
        self.network.send(msg)

    def process_tx_msg(self):
        """Process (send) message in tx queue called from the loop()"""
        cnt = 0
        while self.__txFifo:
            temp = self.__txFifo.popleft()
            self.send_msg(temp)
            # try to limit rate
            cnt = cnt + 1
            if cnt > config.queue_size:
                time.sleep(0.2)
                break

    #####################################################
    # xAAL messages subscribers
    #####################################################
    def receive_msg(self) -> Optional['Message']:
        """return new received message or None"""
        result = None
        data = self.network.get_data()
        if data:
            try:
                msg = self.msg_factory.decode_msg(data, self.msg_filter)
            except MessageParserError as e:
                logger.warning(e)
                msg = None
            result = msg
        return result

    def process_subscribers(self):
        """process incomming messages"""
        msg = self.receive_msg()
        if msg:
            for func in self.subscribers:
                func(msg)
            self.process_attributes_change()

    def handle_request(self, msg: 'Message'):
        """
        Filter msg for devices according default xAAL API then process the
        request for each targets identied in the engine
        """
        if not msg.is_request():
            return

        targets = core.filter_msg_for_devices(msg, self.devices)
        for target in targets:
            if msg.is_request_isalive():
                self.send_alive(target)
            else:
                self.handle_action_request(msg, target)

    def handle_action_request(self, msg: 'Message', target: 'Device'):
        """
        Run method (xAAL exposed method) on device:
            - result is returned if device method gives a response
            - Errors are raised if an error occured:
                * Internal error
                * error returned on the xAAL bus
        """
        if msg.action is None:
            return  # should not happen, but pyright need this check

        try:
            result = run_action(msg, target)
            if result is not None:
                self.send_reply(dev=target, targets=[msg.source], action=msg.action, body=result)
        except CallbackError as e:
            self.send_error(target, e.code, e.description)
        except XAALError as e:
            logger.error(e)

    #####################################################
    # timers
    #####################################################
    def process_timers(self):
        """Process all timers to find out which ones should be run"""
        expire_list = []

        if len(self.timers) != 0:
            now = time.time()
            # little hack to avoid to check timer to often.
            # w/ this enable timer precision is bad, but far enougth
            if (now - self.__last_timer) < 0.4:
                return

            for t in self.timers:
                if t.deadline < now:
                    try:
                        t.func()
                    except CallbackError as e:
                        logger.error(e.description)
                    if t.counter != -1:
                        t.counter -= 1
                        if t.counter == 0:
                            expire_list.append(t)
                    t.deadline = now + t.period
            # delete expired timers
            for t in expire_list:
                self.remove_timer(t)

            self.__last_timer = now

    #####################################################
    # Mainloops & run ..
    #####################################################
    def loop(self):
        """
        Process incomming xAAL msg
        Process timers
        Process attributes change for devices
        Process is_alive for device
        Send msgs from the Tx Buffer
        """
        # Process xAAL msg received, filter msg and process request
        self.process_subscribers()
        # Process timers
        self.process_timers()
        # Process attributes change for devices due to timers
        self.process_attributes_change()
        # Process Alives
        self.process_alives()
        # Process xAAL msgs to send
        self.process_tx_msg()

    def start(self):
        """Start the core engine: send queue alive msg"""
        if self.state in [EngineState.started, EngineState.running]:
            return
        self.network.connect()
        for dev in self.devices:
            self.send_alive(dev)
        self.state = EngineState.started

    def stop(self):
        self.state = EngineState.halted

    def shutdown(self):
        self.stop()

    def run(self):
        self.start()
        self.state = EngineState.running
        while self.state == EngineState.running:
            self.loop()

    def is_running(self) -> bool:
        if self.state == EngineState.running:
            return True
        return False


def run_action(msg: 'Message', device: 'Device') -> Optional[dict]:
    """
    Extract an action & launch it
    Return:
        - action result
        - None if no result

    Note: If an exception raised, it's logged, and raise an XAALError.
    """
    method, params = core.search_action(msg, device)
    result = None
    try:
        result = method(**params)
    except Exception as e:
        logger.error(e)
        raise XAALError("Error in method:%s params:%s" % (msg.action, params))
    # Here result should be None or a dict, and we need to enforce that. This will cause issue
    # in send_reply otherwise.
    if result is not None and not isinstance(result, dict):
        raise XAALError("Method %s should return a dict or None" % msg.action)
    return result





import xaal.lib
from xaal.lib import tools


import logging


ADDR='b8bec7ca-f955-11e6-9031-82ed25e6aaaa'
ADDR=tools.get_random_uuid()

def dump_hex(name,data):
    print("%s : " % name,end='=')
    for k in data:
        if type(k) == int:
            print("%x"%k,end=' ')
        else:
            print("%x"%ord(k),end=' ')
    print()

def test_pysodium():
    from xaal.lib import messages
    import pysodium

    payload = 'FooBar'.encode("utf-8")
    ad = '[]'

    data  = messages.build_timestamp()
    nonce = messages.build_nonce(data)
    key = tools.pass2key("My Friend Goo")

    dump_hex('Payload',payload)
    dump_hex('Key',key)

    ciph  = pysodium.crypto_aead_chacha20poly1305_encrypt(payload, ad, nonce, key)
    dump_hex('Ciph',ciph)

    pjson = pysodium.crypto_aead_chacha20poly1305_decrypt(ciph, ad, nonce, key)
    print(pjson)


def test_device():
    from xaal.lib.devices import Device
    addr = ADDR
    dev = Device("foo.basic",addr)
    dev.vendor_id = 'ACME'
    dev.url="http://acmefactory.blogspot.fr/"
    dev.hw_id = '0x201'
    print(dev.getDescription())
    return dev

def test_msg():
    from xaal.lib.messages import Message
    msg = Message()
    addr = ADDR
    msg.targets = [addr,]
    msg.dump()

def test_encode():
    from xaal.lib.messages import MessageFactory
    key = tools.pass2key('FooBar')
    factory = MessageFactory(key)
    m2 = factory.build_msg()
    print(factory.decode_msg(m2))

def test_log():
    logger = logging.getLogger(__name__)
    logger.info("This is an INFO msg")
    logger.debug("This is an DEBUG msg")


def test_engine():
    from xaal.lib.engine import Engine
    engine = Engine()
    engine.run()

def test_alive():
    addr = ADDR
    dev = xaal.lib.Device("test.basic",addr)
    eng = xaal.lib.Engine()
    eng.add_devices([dev,])
    eng.run()

def test_crypto_decoding_error():
    addr = ADDR
    dev = xaal.lib.Device("test.basic",addr)
    eng = xaal.lib.Engine()
    eng.add_devices([dev,])
    eng.msg_factory.cipher_key = tools.pass2key('FakeKey')
    eng.start()
    eng.loop()


def test_attr():
    dev = xaal.lib.Device("test.basic",ADDR)
    dev.url = "http://linux.org"
    dev.vendor_id = 'ACME'
    dev.product_id = "Full Fake Device"
    dev.info = 'FooBar'
    dev.hw_id = 'ffd0001'
    #dev.alive_period = 100

    attr0 = dev.new_attribute('attr0',10)
    attr1 = dev.new_attribute('attr1',False)

    eng = xaal.lib.Engine()
    eng.add_devices([dev,])

    def update():
        attr0.value = attr0.value + 1

    eng.add_timer(update,60)

    #eng.loop()
    eng.run()



def run():

    logger = tools.get_logger(__name__,logging.DEBUG,"%s.log" % __name__)
    #test_pysodium()
    #test_device()
    #test_msg()
    #test_encode()
    #test_log()
    #test_alive()
    #test_crypto_decoding_error()
    test_attr()



if __name__ == '__main__':
    try:
        run()
    except KeyboardInterrupt:
        print("Bye bye...")



#
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy,
#  Telecom Bretagne.
#
#  This file is part of xAAL.
#
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#

import logging
import time
import typing
from typing import Any, Optional, Union, Callable, Awaitable, Dict

from tabulate import tabulate

from . import bindings, config, tools
from .exceptions import DeviceError

if typing.TYPE_CHECKING:
    from .aioengine import AsyncEngine
    from .engine import Engine
    from .core import EngineMixin

# Funtion types with any arguments and return a dict or None (device methods)
MethodT = Union[Callable[..., Union[dict, None]], Callable[..., Awaitable[Union[dict, None]]]]

logger = logging.getLogger(__name__)


class Attribute(object):
    def __init__(self, name, dev: Optional['Device'] = None, default: Any = None):
        self.name = name
        self.default = default
        self.device: Optional['Device'] = dev
        self.__value = default

    @property
    def value(self) -> Any:
        return self.__value

    @value.setter
    def value(self, value: Any):
        if value != self.__value and self.device:
            eng = self.device.engine
            if eng:
                eng.add_attributes_change(self)
                logger.debug("Attr change %s %s=%s" % (self.device.address, self.name, value))
        self.__value = value

    def __repr__(self) -> str:  # pragma: no cover
        return f"<{self.__module__}.Attribute {self.name} at 0x{id(self):x}>"


class Attributes(list):
    """Devices owns a attributes list. This list also have dict-like access"""

    def __getitem__(self, value):
        if isinstance(value, int):
            return list.__getitem__(self, value)
        for k in self:
            if value == k.name:
                return k.value
        raise KeyError(value)

    def __setitem__(self, name, value):
        if isinstance(name, int):
            return list.__setitem__(self, name, value)
        for k in self:
            if name == k.name:
                k.value = value
                return
        raise KeyError(name)


class Device(object):
    __slots__ = [
        '__dev_type',
        '__address',
        'group_id',
        'vendor_id',
        'product_id',
        'hw_id',
        '__version',
        '__url',
        'schema',
        'info',
        'unsupported_attributes',
        'unsupported_methods',
        'unsupported_notifications',
        'alive_period',
        'next_alive',
        '__attributes',
        'methods',
        'engine',
    ]

    def __init__(
        self,
        dev_type: str,
        addr: Optional[bindings.UUID] = None,
        engine: Union['AsyncEngine', 'Engine', 'EngineMixin', None] = None,
    ):
        # xAAL internal attributes for a device
        self.dev_type = dev_type  # xaal dev_type
        self.address = addr  # xaal addr
        self.group_id: Optional[bindings.UUID] = None  # group devices
        self.vendor_id: Optional[str] = None  # vendor ID ie : ACME
        self.product_id: Optional[str] = None  # product ID
        self.hw_id: Optional[str] = None  # hardware info
        self.version = None  # product release
        self.url = None  # product URL
        self.schema: Optional[str] = None  # schema URL
        self.info: Optional[str] = None  # additionnal info

        # Unsupported stuffs
        self.unsupported_attributes = []
        self.unsupported_methods = []
        self.unsupported_notifications = []
        # Alive management
        self.alive_period = config.alive_timer  # time in sec between two alive
        self.next_alive = 0
        # Default attributes & methods
        self.__attributes = Attributes()
        self.methods: Dict[str, MethodT] = {
            'get_attributes': self._get_attributes,
            'get_description': self._get_description,
        }
        self.engine = engine

    @property
    def dev_type(self) -> str:
        return self.__dev_type

    @dev_type.setter
    def dev_type(self, value: str):
        if not tools.is_valid_dev_type(value):
            raise DeviceError(f"The dev_type {value} is not valid")
        self.__dev_type = value

    @property
    def version(self) -> Optional[str]:
        return self.__version

    @version.setter
    def version(self, value: Any):
        # version must be a string
        if value:
            self.__version = "%s" % value
        else:
            self.__version = None

    @property
    def address(self) -> Optional[bindings.UUID]:
        return self.__address

    @address.setter
    def address(self, value: Optional[bindings.UUID]):
        if value is None:
            self.__address = None
            return
        if not tools.is_valid_address(value):
            raise DeviceError("This address is not valid")
        self.__address = value

    @property
    def url(self) -> Optional[bindings.URL]:
        return self.__url

    @url.setter
    def url(self, value: Optional[str]):
        if value is None:
            self.__url = None
        else:
            self.__url = bindings.URL(value)

    # attributes
    def new_attribute(self, name: str, default: Any = None):
        attr = Attribute(name, self, default)
        self.add_attribute(attr)
        return attr

    def add_attribute(self, attr: Attribute):
        if attr:
            self.__attributes.append(attr)
            attr.device = self

    def del_attribute(self, attr: Attribute):
        if attr:
            attr.device = None
            self.__attributes.remove(attr)

    def get_attribute(self, name: str) -> Optional[Attribute]:
        for attr in self.__attributes:
            if attr.name == name:
                return attr
        return None

    @property
    def attributes(self) -> Attributes:
        return self.__attributes

    @attributes.setter
    def attributes(self, values: Attributes):
        if isinstance(values, Attributes):
            self.__attributes = values
        else:
            raise DeviceError("Invalid attributes list, use class Attributes)")

    def add_method(self, name: str, func: MethodT):
        self.methods.update({name: func})

    def del_method(self, name: str):
        if name in self.methods:
            del self.methods[name]

    def get_methods(self) -> Dict[str, MethodT]:
        return self.methods

    def update_alive(self):
        """update the alive timimg"""
        self.next_alive = time.time() + self.alive_period

    def get_timeout(self) -> int:
        """return Alive timeout used for isAlive msg"""
        return 2 * self.alive_period

    #####################################################
    # Usefull methods
    #####################################################
    def dump(self):
        print("= Device: %s" % self)
        # info & description
        r = []
        r.append(['dev_type', self.dev_type])
        r.append(['address', self.address])
        for k, v in self._get_description().items():
            r.append([k, v])
        print(tabulate(r, tablefmt='psql'))

        # attributes
        if len(self._get_attributes()) > 0:
            r = []
            for k, v in self._get_attributes().items():
                r.append([k, str(v)])
            print(tabulate(r, tablefmt='psql'))

        # methods
        if len(self.methods) > 0:
            r = []
            for k, v in self.methods.items():
                r.append([k, v.__name__])
            print(tabulate(r, tablefmt='psql'))

    def __repr__(self) -> str:
        return f"<xaal.Device {id(self):x} {self.address} {self.dev_type}>"

    #####################################################
    # default public methods
    #####################################################
    def _get_description(self) -> dict:
        result = {}
        if self.vendor_id:
            result['vendor_id'] = self.vendor_id
        if self.product_id:
            result['product_id'] = self.product_id
        if self.version:
            result['version'] = self.version
        if self.url:
            result['url'] = self.url
        if self.schema:
            result['schema'] = self.schema
        if self.info:
            result['info'] = self.info
        if self.hw_id:
            result['hw_id'] = self.hw_id
        if self.group_id:
            result['group_id'] = self.group_id
        if self.unsupported_methods:
            result['unsupported_methods'] = self.unsupported_methods
        if self.unsupported_notifications:
            result['unsupported_notifications'] = self.unsupported_notifications
        if self.unsupported_attributes:
            result['unsupported_attributes'] = self.unsupported_attributes
        return result

    def _get_attributes(self, _attributes=None):
        """
        attributes:
            - None = body empty and means request all attributes
            - Empty array means request all attributes
            - Array of attributes (string) and means request attributes in the
              list

        TODO: (Waiting for spec. decision) add test on attribute devices
            - case physical sensor not responding or value not ready add error
            with specific error code and with value = suspicious/stale/cached
        """
        result = {}
        dev_attr = {attr.name: attr for attr in self.__attributes}
        if _attributes:
            # Process attributes filter
            for attr in _attributes:
                if attr in dev_attr.keys():
                    result.update({dev_attr[attr].name: dev_attr[attr].value})
                else:
                    logger.debug(f"Attribute {attr} not found")
        else:
            # Process all attributes
            for attr in dev_attr.values():
                result.update({attr.name: attr.value})
        return result

    def send_notification(self, notification: str, body: dict = {}):
        """queue an notification, this is just a method helper"""
        if self.engine:
            self.engine.send_notification(self, notification, body)



# Load main class & modules.


from .config import config
from . import tools
from . import bindings
from . import aiohelpers as helpers

from .core import Timer

# sync engine
from .engine import Engine
from .network import NetworkConnector

# async engine
from .aioengine import AsyncEngine
from .aionetwork import AsyncNetworkConnector

from .devices import Device, Attribute, Attributes
from .messages import Message, MessageFactory, MessageType, MessageAction
from .exceptions import *



import asyncio
import logging
import signal
import sys
import time
import typing
from enum import Enum
from pprint import pprint
from typing import Any, Optional
from uuid import UUID

import aioconsole
from tabulate import tabulate

from .config import config
from . import core
from .aionetwork import AsyncNetworkConnector
from .exceptions import CallbackError, XAALError
from .messages import MessageParserError

if typing.TYPE_CHECKING:
    from .devices import Device
    from .messages import Message

logger = logging.getLogger(__name__)


#####################################################
# Hooks
#####################################################
class HookType(Enum):
    start = 0
    stop = 1


class Hook(object):
    __slots__ = ['type', 'func', 'args', 'kwargs']

    def __init__(self, type_: HookType, func: core.FuncT, *args, **kwargs):
        # func has to be a callable, but it can be a coroutine or a function
        self.type = type_
        self.func = func
        self.args = args
        self.kwargs = kwargs


class AsyncEngine(core.EngineMixin):
    __slots__ = [
        '__txFifo',
        '_loop',
        '_tasks',
        '_hooks',
        '_watchdog_task',
        '_kill_counter',
        'running_event',
        'watchdog_event',
        'started_event',
    ]

    def __init__(
        self, address: str = config.address, port: int = config.port, hops: int = config.hops, key: bytes = config.key
    ):
        core.EngineMixin.__init__(self, address, port, hops, key)

        self.__txFifo = asyncio.Queue()  # tx msg fifo
        self._loop = None  # event loop
        self._hooks = []  # hooks
        self._tasks = []  # tasks
        self._watchdog_task = None  # watchdog task
        self._kill_counter = 0  # watchdog counter

        self.started_event = asyncio.Event()  # engine started event
        self.running_event = asyncio.Event()  # engine running event
        self.watchdog_event = asyncio.Event()  # watchdog event

        signal.signal(signal.SIGTERM, self.sigkill_handler)
        signal.signal(signal.SIGINT, self.sigkill_handler)

        # message receive workflow
        self.subscribe(self.handle_request)
        # start network
        self.network = AsyncNetworkConnector(address, port, hops)

    #####################################################
    # Hooks
    #####################################################
    def on_start(self, func: core.FuncT, *args, **kwargs):
        hook = Hook(HookType.start, func, *args, **kwargs)
        self._hooks.append(hook)

    def on_stop(self, func: core.FuncT, *args, **kwargs):
        hook = Hook(HookType.stop, func, *args, **kwargs)
        self._hooks.append(hook)

    async def run_hooks(self, hook_type: HookType):
        hooks = list(filter(lambda hook: hook.type == hook_type, self._hooks))
        if len(hooks) != 0:
            logger.debug(f"Launching {hook_type} hooks")
            for h in hooks:
                await run_func(h.func, *h.args, **h.kwargs)

    #####################################################
    # timers
    #####################################################
    async def process_timers(self):
        """Process all timers to find out which ones should be run"""
        expire_list = []
        now = time.time()
        for t in self.timers:
            if t.deadline < now:
                try:
                    await run_func(t.func)
                except CallbackError as e:
                    logger.error(e.description)
                if t.counter != -1:
                    t.counter -= 1
                    if t.counter == 0:
                        expire_list.append(t)
                t.deadline = now + t.period
        # delete expired timers
        for t in expire_list:
            self.remove_timer(t)

    #####################################################
    # msg send / receive
    #####################################################
    def queue_msg(self, msg: bytes):
        """queue a message"""
        self.__txFifo.put_nowait(msg)

    def send_msg(self, msg: bytes):
        """Send an encoded message to the bus, use queue_msg instead"""
        self.network.send(msg)

    async def receive_msg(self) -> Optional['Message']:
        """return new received message or None"""
        data = await self.network.get_data()
        if data:
            try:
                msg = self.msg_factory.decode_msg(data, self.msg_filter)
            except MessageParserError as e:
                logger.warning(e)
                msg = None
            return msg
        return None

    async def process_rx_msg(self):
        """process incomming messages"""
        msg = await self.receive_msg()
        if msg:
            for func in self.subscribers:
                await run_func(func, msg)
            self.process_attributes_change()

    def handle_request(self, msg: 'Message'):
        """Filter msg for devices according default xAAL API then process the
        request for each targets identied in the engine
        """
        if not msg.is_request():
            return

        targets = core.filter_msg_for_devices(msg, self.devices)
        for target in targets:
            if msg.is_request_isalive():
                self.send_alive(target)
            else:
                self.new_task(self.handle_action_request(msg, target))

    async def handle_action_request(self, msg: 'Message', target: 'Device'):
        if msg.action is None:
            return  # should not happen, but pyright need this check

        try:
            result = await run_action(msg, target)
            if result is not None and type(result) is dict:
                self.send_reply(dev=target, targets=[msg.source], action=msg.action, body=result)
        except CallbackError as e:
            self.send_error(target, e.code, e.description)
        except XAALError as e:
            logger.error(e)

    #####################################################
    # Asyncio loop & Tasks
    #####################################################
    def get_loop(self) -> asyncio.AbstractEventLoop:
        if self._loop is None:
            try:
                self._loop = asyncio.get_running_loop()
            except RuntimeError:
                logger.debug("New event loop")
                self._loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self._loop)
        return self._loop

    def new_task(self, coro: Any, name: Optional[str] = None) -> asyncio.Task:
        # we maintain a task list, to be able to stop/start the engine
        # on demand. needed by HASS
        task = self.get_loop().create_task(coro, name=name)
        self._tasks.append(task)
        task.add_done_callback(self.task_callback)
        return task

    def task_callback(self, task: asyncio.Task):
        # called when a task ended
        self._tasks.remove(task)

    def all_tasks(self) -> typing.List[asyncio.Task]:
        return self._tasks

    async def boot_task(self):
        self.watchdog_event.clear()
        # queue the alive before anything
        for dev in self.devices:
            self.send_alive(dev)
        await self.network.connect()
        self.running_event.set()
        await self.run_hooks(HookType.start)

    async def receive_task(self):
        await self.running_event.wait()
        while self.is_running():
            await self.process_rx_msg()

    async def send_task(self):
        await self.running_event.wait()
        while self.is_running():
            temp = await self.__txFifo.get()
            self.send_msg(temp)

    async def timer_task(self):
        await self.running_event.wait()
        self.setup_alives_timer()
        while self.is_running():
            await asyncio.sleep(0.2)
            await self.process_timers()
            self.process_attributes_change()

    async def watchdog_task(self):
        await self.watchdog_event.wait()
        await self.stop()
        logger.info("Exit")

    #####################################################
    # start / stop / shutdown
    #####################################################
    def is_running(self) -> bool:
        return self.running_event.is_set()

    def start(self):
        if self.is_running():
            logger.warning("Engine already started")
            return
        self.started_event.set()
        self.new_task(self.boot_task(), name='Boot')
        self.new_task(self.receive_task(), name='RecvQ')
        self.new_task(self.send_task(), name='SendQ')
        self.new_task(self.timer_task(), name='Timers')
        if config.remote_console:
            self.new_task(console(locals()), name='Console')

    def setup_alives_timer(self):
        # needed on stop-start sequence
        if self.process_alives in [t.func for t in self.timers]:
            return
        # process alives every 10 seconds
        self.add_timer(self.process_alives, 10)

    async def stop(self):  # pyright: ignore
        logger.info("Stopping engine")
        await self.run_hooks(HookType.stop)
        self.running_event.clear()
        self.started_event.clear()
        # cancel all tasks
        for task in self.all_tasks():
            if task != self._watchdog_task and not task.cancelled():
                task.cancel()

    def sigkill_handler(self, signal, frame):
        print("", end="\r")  # remove the uggly ^C
        if not self.is_running():
            logger.warning("Engine already stopped")
            self._kill_counter = 1
        self._kill_counter += 1
        self.shutdown()
        if self._kill_counter > 1:
            logger.warning("Force quit")
            sys.exit(-1)
        else:
            logger.warning("Kill requested")

    def shutdown(self):
        self.watchdog_event.set()

    def run(self):
        if not self.started_event.is_set():
            self.start()
        if self._watchdog_task is None:
            # start the watchdog task
            self._watchdog_task = self.new_task(self.watchdog_task(), name="Watchdog task")
            self.get_loop().run_until_complete(self._watchdog_task)
        else:
            logger.warning("Engine already running")

    #####################################################
    # Debugging tools
    #####################################################
    def dump_timers(self):
        headers = ['Func', 'Period', 'Counter', 'Deadline']
        rows = []
        now = time.time()
        for t in self.timers:
            remain = round(t.deadline - now, 1)
            rows.append([str(t.func), t.period, t.counter, remain])
        print("= Timers")
        print(tabulate(rows, headers=headers, tablefmt='fancy_grid'))

    def dump_tasks(self):
        headers = ['Name', 'Coro', 'Loop ID']
        rows = []
        for t in self.all_tasks():
            rows.append([t.get_name(), str(t.get_coro()), id(t.get_loop())])
        print("= Tasks")
        print(tabulate(rows, headers=headers, tablefmt='fancy_grid'))

    def dump_devices(self):
        headers = ['addr', 'dev_type', 'info']
        rows = []
        for d in self.devices:
            rows.append([d.address, d.dev_type, d.info])
        print("= Devices")
        print(tabulate(rows, headers=headers, tablefmt='fancy_grid'))

    def dump_hooks(self):
        headers = ['Type', 'Hook']
        rows = []
        for h in self._hooks:
            rows.append([h.type, str(h.func)])
        print("= Hooks")
        print(tabulate(rows, headers=headers, tablefmt='fancy_grid'))

    def dump(self):
        self.dump_devices()
        self.dump_tasks()
        self.dump_timers()
        self.dump_hooks()

    def get_device(self, uuid: UUID) -> Optional['Device']:
        # TODO:Check if this method is usefull
        for dev in self.devices:
            if dev.address == uuid:
                return dev
        return None


#####################################################
# Utilities functions
#####################################################
async def run_func(func, *args, **kwargs):
    """run a function or a coroutine function"""
    if asyncio.iscoroutinefunction(func):
        return await func(*args, **kwargs)
    else:
        return func(*args, **kwargs)


async def run_action(msg: 'Message', device: 'Device'):
    """
    Extract an action & launch it
    Return:
        - action result
        - None if no result

    Notes:
        - If an exception raised, it's logged, and raise an XAALError.
        - Same API as legacy Engine, but accept coroutine functions
    """
    method, params = core.search_action(msg, device)
    result = None
    try:
        if asyncio.iscoroutinefunction(method):
            result = await method(**params)
        else:
            result = method(**params)
    except Exception as e:
        logger.error(e)
        raise XAALError("Error in method:%s params:%s" % (msg.action, params))
    return result


#####################################################
# Debugging console
#####################################################
async def console(locals=locals(), port: Optional[int] = None):
    """launch a console to enable remote engine inspection"""
    if port is None:
        # let's find a free port if not specified
        def find_free_port():
            import socketserver

            with socketserver.TCPServer(('localhost', 0), None) as s:  # pyright: ignore pyright reject the None here
                return s.server_address[1]

        port = find_free_port()

    logger.debug(f"starting debug console on port {port}")
    sys.ps1 = "[xAAL] >>> "
    banner = "=" * 78 + "\nxAAL remote console\n" + "=" * 78
    locals.update({'pprint': pprint})

    def factory(streams):
        return aioconsole.AsynchronousConsole(locals=locals, streams=streams)

    # start the console
    try:
        # debian with ipv6 disabled still state that localhost is ::1, which broke aioconsole
        await aioconsole.start_interactive_server(
            host="127.0.0.1", port=port, factory=factory, banner=banner
        )  # pyright: ignore
    except OSError:
        logger.warning("Unable to run console")



"""
This file contains some helpers functions. This functions aren't used in the lib itself
but can be usefull for xaal packages developpers
"""

import logging
import logging.handlers
import os
import time
from typing import Any, Optional, Union

import coloredlogs
from decorator import decorator

from .config import config


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


@decorator
def timeit(method, *args, **kwargs):
    logger = logging.getLogger(__name__)
    ts = time.time()
    result = method(*args, **kwargs)
    te = time.time()
    logger.debug("%r (%r, %r) %2.6f sec" % (method.__name__, args, kwargs, te - ts))
    return result


def set_console_title(value: str):
    # set xterm title
    print("\x1b]0;xAAL => %s\x07" % value, end="\r")


def setup_console_logger(level: Union[str, int] = config.log_level):
    # fmt = "%(asctime)s %(name)-25s %(funcName)-18s %(levelname)-8s %(message)s"
    fmt = "%(asctime)s %(name)-25s %(funcName)-18s %(message)s"
    # fmt = '[%(name)s] %(funcName)s %(levelname)s: %(message)s'
    coloredlogs.install(level=level, fmt=fmt)


def setup_file_logger(name: str, level: Union[str, int] = config.log_level, filename: Optional[str] = None):
    filename = filename or os.path.join(config.log_path, "%s.log" % name)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
    handler = logging.handlers.RotatingFileHandler(filename, "a", 10000, 1, "utf8")
    handler.setLevel(level)
    handler.setFormatter(formatter)
    # register the new handler
    logger = logging.getLogger(name)
    logger.root.addHandler(handler)
    logger.root.setLevel('DEBUG')


# ---------------------------------------------------------------------------
# TBD: We should merge this stuffs, and add support for default config file
#      and commnand line parsing.
#
# Default arguments console_log and file_log are (and should) never be used.
# ---------------------------------------------------------------------------
def run_package(pkg_name: str, pkg_setup: Any, console_log: bool = True, file_log: bool = False):
    if console_log:
        set_console_title(pkg_name)
        setup_console_logger()
    if file_log:
        setup_file_logger(pkg_name)
    logger = logging.getLogger(pkg_name)
    logger.info("starting xaal package: %s" % pkg_name)

    from .engine import Engine

    eng = Engine()
    result = pkg_setup(eng)

    if result is not True:
        logger.critical("something goes wrong with package: %s" % pkg_name)
    try:
        eng.run()
    except KeyboardInterrupt:
        eng.shutdown()
        logger.info("exit")


__all__ = ['singleton', 'timeit', 'set_console_title', 'setup_console_logger', 'setup_file_logger', 'run_package']



#
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy,
#  Telecom Bretagne.
#
#  This file is part of xAAL.
#
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#

import functools
import os
import re
import sys
from typing import Optional, Union

import pysodium
from configobj import ConfigObj

from .config import config
from .bindings import UUID

XAAL_DEVTYPE_PATTERN = "^[a-zA-Z][a-zA-Z0-9_-]*\\.[a-zA-Z][a-zA-Z0-9_-]*$"


def get_cfg_filename(name: str, cfg_dir: str = config.conf_dir) -> str:
    if name.startswith("xaal."):
        name = name[5:]
    filename = "%s.ini" % name
    if not os.path.isdir(cfg_dir):
        print("Your configuration directory doesn't exist: [%s]" % cfg_dir)
    return os.path.join(cfg_dir, filename)


def load_cfg_file(filename: str) -> Optional[ConfigObj]:
    """load .ini file and return it as dict"""
    if os.path.isfile(filename):
        return ConfigObj(filename, indent_type="  ", encoding='utf8')
    return None


def load_cfg(app_name: str) -> Optional[ConfigObj]:
    filename = get_cfg_filename(app_name)
    return load_cfg_file(filename)


def load_cfg_or_die(app_name: str) -> ConfigObj:
    cfg = load_cfg(app_name)
    if not cfg:
        print("Unable to load config file %s" % get_cfg_filename(app_name))
        sys.exit(-1)
    return cfg


def new_cfg(app_name: str) -> ConfigObj:
    filename = get_cfg_filename(app_name)
    cfg = ConfigObj(filename, indent_type="  ")
    cfg['config'] = {'addr' : get_random_uuid().str}
    return cfg


def get_random_uuid() -> UUID:
    return UUID.random()


def get_random_base_uuid(digit=2) -> UUID:
    return UUID.random_base(digit)


def get_uuid(val: Union[UUID, str]) -> Optional[UUID]:
    if isinstance(val, UUID):
        return val
    if isinstance(val, str):
        return str_to_uuid(val)
    return None


def str_to_uuid(val: str) -> Optional[UUID]:
    """return an xAAL address for a given string"""
    try:
        return UUID(val)
    except ValueError:
        return None


def bytes_to_uuid(val: bytes) -> Optional[UUID]:
    try:
        return UUID(bytes=val)
    except ValueError:
        return None


def is_valid_uuid(val: UUID) -> bool:
    return isinstance(val, UUID)


def is_valid_address(val: UUID) -> bool:
    return is_valid_uuid(val)


@functools.lru_cache(maxsize=128)
def is_valid_dev_type(val: str) -> bool:
    if not isinstance(val, str):
        return False
    if re.match(XAAL_DEVTYPE_PATTERN, val):
        return True
    return False


def pass2key(passphrase: str) -> bytes:
    """Generate key from passphrase using libsodium
    crypto_pwhash_scryptsalsa208sha256 func
    salt: buffer of zeros
    opslimit: crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
    memlimit: crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE
    """
    buf = passphrase.encode("utf-8")
    KEY_BYTES = pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES  # 32
    # this should be:
    # salt = bytes(KEY_BYTES)
    # but due to bytes() stupid stuff in py2 we need this awfull stuff
    salt = ('\00' * KEY_BYTES).encode('utf-8')
    opslimit = pysodium.crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
    memlimit = pysodium.crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE
    key = pysodium.crypto_pwhash_scryptsalsa208sha256(KEY_BYTES, buf, salt, opslimit, memlimit)
    return key


@functools.lru_cache(maxsize=128)
def reduce_addr(addr: UUID) -> str:
    """return a string based addred without all  digits"""
    tmp = addr.str
    return tmp[:5] + '..' + tmp[-5:]



import unittest
from xaal.lib import cbor
from xaal.lib import bindings
import binascii


class TestCbor(unittest.TestCase):
    def test_encode(self):
        value = cbor.dumps(42)
        self.assertEqual(value, b"\x18*")  # 0x18 0x2a

    def test_decode(self):
        value = cbor.loads(b"\x18*")
        self.assertEqual(value, 42)

    def test_list(self):
        value = []
        value.append(7)
        value.append(1234567890)
        data = cbor.dumps(value)
        print(data)
        self.assertEqual(data, binascii.unhexlify(b"82071a499602d2"))

    def test_uuid(self):
        uuid = bindings.UUID.random()
        cbor_value = cbor.dumps(uuid)
        value = cbor.loads(cbor_value)
        self.assertEqual(value, uuid)
        self.assertEqual(type(value), bindings.UUID)

    def test_url(self):
        url = bindings.URL("http://www.example.com")
        cbor_value = cbor.dumps(url)
        value = cbor.loads(cbor_value)
        self.assertEqual(value, url)
        self.assertEqual(type(value), bindings.URL)

    def test_cleanup(self):
        data = [
            7,
            "hello",
            bindings.UUID("12345678-1234-1234-1234-123456789012"),
            bindings.URL("http://www.example.com"),
            {"hello": "world"},
        ]
        cbor_value = cbor.dumps(data)
        value = cbor.loads(cbor_value)
        cbor.cleanup(value)
        self.assertEqual(
            value,
            [
                7,
                "hello",
                "12345678-1234-1234-1234-123456789012",
                "http://www.example.com",
                {"hello": "world"},
            ],
        )


if __name__ == "__main__":
    unittest.main()



import inspect
import logging
import time
import typing
from typing import Any, Awaitable, Optional, List, Callable, Union

from .exceptions import EngineError, XAALError
from .messages import ALIVE_ADDR, MessageAction, MessageFactory, MessageType

if typing.TYPE_CHECKING:
    from .devices import Device, Attribute
    from .messages import Message

# Function type w/ no argument, and no return (Timer, Hooks)
FuncT = Union[Callable[[], None], Callable[[], Awaitable[None]]]

# Function type w/ message as argument (subscribers), no return
SubFuncT = Union[Callable[['Message'], None], Callable[['Message'], Awaitable[None]]]

logger = logging.getLogger(__name__)


#####################################################
# Timer class
#####################################################
class Timer(object):
    def __init__(self, func: FuncT, period: int, counter: int):
        # Timer function should a Callable[[],None], but it can be a coroutine too
        self.func = func
        self.period = period
        self.counter = counter
        self.deadline = time.time() + period


class EngineMixin(object):
    __slots__ = ['devices', 'timers', 'subscribers', 'msg_filter', '_attributesChange', 'network', 'msg_factory']

    def __init__(self, address: str, port: int, hops: int, key: bytes):
        self.devices: list[Device] = []  # list of devices / use (un)register_devices()
        self.timers: list[Timer] = []  # functions to call periodic
        self.subscribers: list[SubFuncT] = []  # message receive workflow
        self.msg_filter = None  # message filter

        self._attributesChange = []  # list of XAALAttributes instances

        # network connector
        self.network = None
        # start msg worker
        self.msg_factory = MessageFactory(key)
        # filter function activated
        self.enable_msg_filter()

    #####################################################
    # Devices management
    #####################################################
    def add_device(self, dev: 'Device'):
        """register a new device"""
        if dev not in self.devices:
            self.devices.append(dev)
            dev.engine = self
        if self.is_running():
            self.send_alive(dev)

    def add_devices(self, devs: List['Device']):
        """register new devices"""
        for dev in devs:
            self.add_device(dev)

    def remove_device(self, dev: 'Device'):
        """unregister a device"""
        dev.engine = None
        # Remove dev from devices list
        self.devices.remove(dev)

    #####################################################
    # xAAL messages Tx handling
    #####################################################
    # Fifo for msg to send
    def queue_msg(self, msg: bytes):
        logger.critical("To be implemented queue_msg: %s", msg)

    def send_request(self, dev: 'Device', targets: list, action: str, body: Optional[dict] = None):
        """queue a new request"""
        msg = self.msg_factory.build_msg(dev, targets, MessageType.REQUEST, action, body)
        self.queue_msg(msg)

    def send_reply(self, dev: 'Device', targets: list, action: str, body: Optional[dict] = None):
        """queue a new reply"""
        msg = self.msg_factory.build_msg(dev, targets, MessageType.REPLY, action, body)
        self.queue_msg(msg)

    def send_error(self, dev: 'Device', errcode: int, description: Optional[str] = None):
        """queue a error message"""
        msg = self.msg_factory.build_error_msg(dev, errcode, description)
        self.queue_msg(msg)

    def send_get_description(self, dev: 'Device', targets: list):
        """queue a get_description request"""
        self.send_request(dev, targets, MessageAction.GET_DESCRIPTION.value)

    def send_get_attributes(self, dev: 'Device', targets: list):
        """queue a get_attributes request"""
        self.send_request(dev, targets, MessageAction.GET_ATTRIBUTES.value)

    def send_notification(self, dev: 'Device', action: str, body: Optional[dict] = None):
        """queue a notificaton"""
        msg = self.msg_factory.build_msg(dev, [], MessageType.NOTIFY, action, body)
        self.queue_msg(msg)

    def send_alive(self, dev: 'Device'):
        """Send a Alive message for a given device"""
        timeout = dev.get_timeout()
        msg = self.msg_factory.build_alive_for(dev, timeout)
        self.queue_msg(msg)
        dev.update_alive()

    def send_is_alive(self, dev: 'Device', targets: list = [ALIVE_ADDR], dev_types: list = ["any.any"]):
        """Send a is_alive message, w/ dev_types filtering"""
        body = {'dev_types': dev_types}
        self.send_request(dev, targets, MessageAction.IS_ALIVE.value, body)

    #####################################################
    # Messages filtering
    #####################################################
    def enable_msg_filter(self, func: Optional[Callable[['Message'], bool]] = None):
        """enable message filter"""
        self.msg_filter = func or self.default_msg_filter

    def disable_msg_filter(self):
        """disable message filter"""
        self.msg_filter = None

    def default_msg_filter(self, msg: 'Message'):
        """
        Filter messages:
        - check if message has alive request address
        - check if the message is for us
        return False, if message should be dropped
        """
        # Alive request
        if ALIVE_ADDR in msg.targets:
            return True
        # Managed device ?
        for dev in self.devices:
            if dev.address in msg.targets:
                return True
        return False

    #####################################################
    # Alive messages
    #####################################################
    def process_alives(self):
        """Periodic sending alive messages"""
        now = time.time()
        for dev in self.devices:
            if dev.next_alive < now:
                self.send_alive(dev)

    #####################################################
    # xAAL attributes changes
    #####################################################
    def add_attributes_change(self, attr: 'Attribute'):
        """add a new attribute change to the list"""
        self._attributesChange.append(attr)

    def get_attributes_change(self) -> List['Attribute']:
        """return the pending attributes changes list"""
        return self._attributesChange

    def process_attributes_change(self):
        """Processes (send notify) attributes changes for all devices"""
        devices = {}
        # Group attributes changed by device
        for attr in self.get_attributes_change():
            if attr.device not in devices.keys():
                devices[attr.device] = {}
            devices[attr.device][attr.name] = attr.value

        for dev in devices:
            self.send_notification(dev, MessageAction.ATTRIBUTES_CHANGE.value, devices[dev])
        self._attributesChange = []  # empty array

    #####################################################
    # xAAL messages subscribers
    #####################################################
    def subscribe(self, func: SubFuncT):
        # func should be a Callable[[Message],None], but it can be a coroutine too
        self.subscribers.append(func)

    def unsubscribe(self, func: SubFuncT):
        self.subscribers.remove(func)

    #####################################################
    # timers
    #####################################################
    def add_timer(self, func: FuncT, period: int, counter: int = -1):
        """
        func: function to call
        period: period in second
        counter: number of repeat, -1 => always
        """
        if counter == 0:
            raise EngineError("Timer counter should =-1 or >0")
        t = Timer(func, period, counter)
        self.timers.append(t)
        return t

    def remove_timer(self, timer: Timer):
        """remove a given timer from the list"""
        self.timers.remove(timer)

    #####################################################
    # start/stop/run API
    #####################################################
    def start(self):
        logger.critical("To be implemented start")

    def stop(self):
        logger.critical("To be implemented stop")

    def shutdown(self):
        logger.critical("To be implemented shutdown")

    def run(self):
        logger.critical("To be implemented run")

    def is_running(self) -> bool:
        logger.critical("To be implemented is_running")
        return False


#####################################################
# Usefull functions to Engine developpers
#####################################################
def filter_msg_for_devices(msg: 'Message', devices: List['Device']) -> List['Device']:
    """
    loop throught the devices, to find which are expected w/ the msg
    - Filter on dev_types for is_alive broadcast request.
    - Filter on device address
    """
    results = []
    if msg.is_request_isalive() and (ALIVE_ADDR in msg.targets):
        # if we receive a broadcast is_alive request, we reply
        # with filtering on dev_tyes.
        if 'dev_types' in msg.body.keys():
            dev_types = msg.body['dev_types']
            if "any.any" in dev_types:
                results = devices
            else:
                for dev in devices:
                    any_subtype = dev.dev_type.split(".")[0] + ".any"
                    if dev.dev_type in dev_types:
                        results.append(dev)
                    elif any_subtype in dev_types:
                        results.append(dev)
    else:
        # this is a normal request, only filter on device address
        # note: direct is_alive are treated like normal request
        # so dev_types filtering is discarded
        for dev in devices:
            if dev.address in msg.targets:
                results.append(dev)
    return results


def search_action(msg: 'Message', device: 'Device'):
    """
    Extract an action (match with methods) from a msg on the device.
    Return:
        - None
        - found method & matching parameters

    Note: If method not found raise error, if wrong parameter error log
    """
    methods = device.get_methods()
    params = {}
    result = None
    if msg.action in methods.keys():
        assert msg.action
        method = methods[msg.action]
        body_params = None
        if msg.body:
            method_params = get_args_method(method)
            body_params = msg.body

            for k in body_params:
                temp = "_%s" % k
                if temp in method_params:
                    params.update({temp: body_params[k]})
                else:
                    logger.warning("Wrong method parameter [%s] for action %s" % (k, msg.action))
        result = (method, params)
    else:
        raise XAALError("Method %s not found on device %s" % (msg.action, device))
    return result


def get_args_method(method: Any) -> List[str]:
    """return the list on arguments for a given python method"""
    spec = inspect.getfullargspec(method)
    try:
        spec.args.remove('self')
    except Exception:
        pass
    return spec.args



from . helpers import *
from decorator import decorator
import asyncio
import logging

@decorator
def spawn(func, *args, **kwargs):
    return asyncio.get_event_loop().run_in_executor(None, func, *args, **kwargs)


def static_vars(**kwargs_):
    def decorate(func):
        for k in kwargs_:
            setattr(func, k, kwargs_[k])
        return func
    return decorate


def run_async_package(pkg_name, pkg_setup, console_log=True, file_log=False):
    if console_log:
        set_console_title(pkg_name)
        setup_console_logger()
    if file_log:
        setup_file_logger(pkg_name)

    from .aioengine import AsyncEngine
    eng = AsyncEngine()
    eng.start()
    logger = logging.getLogger(pkg_name)
    logger.info("starting xaal package: %s"% pkg_name )
    result = pkg_setup(eng)
    if result is not True:
        logger.critical("something goes wrong with package: %s" % pkg_name)
    try:
        eng.run()
    except KeyboardInterrupt:
        eng.shutdown()
        logger.info("Exit")



# Default configuration

import os
import sys
import binascii
from configobj import ConfigObj

# Default settings
DEF_ADDR = "224.0.29.200"
DEF_PORT = 1236
DEF_HOPS = 10
DEF_ALIVE_TIMER = 100
DEF_CIPHER_WINDOW = 60 * 2
DEF_QUEUE_SIZE = 10
DEF_LOG_LEVEL = "DEBUG"
DEF_LOG_PATH = "/var/log/xaal"
DEF_REMOTE_CONSOLE = False


STACK_VERSION = 7


class Config:
    def __init__(self):
        self.conf_dir = os.environ.get("XAAL_CONF_DIR", os.path.expanduser("~/.xaal"))
        self.address = DEF_ADDR
        self.port = DEF_PORT
        self.hops = DEF_HOPS
        self.alive_timer = DEF_ALIVE_TIMER
        self.cipher_window = DEF_CIPHER_WINDOW
        self.queue_size = DEF_QUEUE_SIZE
        self.log_level = DEF_LOG_LEVEL
        self.log_path = DEF_LOG_PATH
        self.remote_console = DEF_REMOTE_CONSOLE
        self.key = b''
        self.STACK_VERSION = STACK_VERSION

    def load(self, name="xaal.ini"):
        filename = os.path.join(self.conf_dir, name)
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"Unable to load xAAL config file [{filename}]")

        cfg = ConfigObj(filename)
        self.address = self.safe_string(cfg.get('address'), DEF_ADDR)
        self.port = self.safe_int(cfg.get('port'), DEF_PORT)
        self.hops = self.safe_int(cfg.get('hops'), DEF_HOPS)
        self.alive_timer = self.safe_int(cfg.get('alive_timer'), DEF_ALIVE_TIMER)
        self.cipher_window = self.safe_int(cfg.get('cipher_window'), DEF_CIPHER_WINDOW)
        self.queue_size = self.safe_int(cfg.get('queue_size'), DEF_QUEUE_SIZE)
        self.log_level = self.safe_string(cfg.get('log_level'), DEF_LOG_LEVEL)
        self.log_path = self.safe_string(cfg.get('log_path'), DEF_LOG_PATH)
        self.remote_console = self.safe_bool(cfg.get('remote_console'), DEF_REMOTE_CONSOLE)

        key = cfg.get('key', None)
        if key and type(key) is str:
            self.key = binascii.unhexlify(key.encode('utf-8'))
        else:
            raise ValueError(f"Key not set in config file [{filename}]")

    ## Helper functions
    # Pylint enforce to sanity check the input. In fact, ConfigObj can do the job without issue
    # but Pytlint assume cfg.get can return None (even w/ default set), so it warm about wrong
    # type in all config setting. By doing this I insure that the value is of the right type.

    @staticmethod
    def safe_int(value, default):
        try:
            return int(value)
        except (ValueError, TypeError):
            return default

    @staticmethod
    def safe_string(value, default):
        if value is None:
            return default
        try:
            return str(value)
        except (ValueError, TypeError):
            return default

    @staticmethod
    def safe_bool(value, default):
        if value is None:
            return default
        if value.lower() == "true":
            return True
        return False


config_dir = os.environ.get("XAAL_CONF_DIR", None)
config = Config()

if config_dir is None or config_dir != '':
    try:
        config.load()
    except Exception as e:
        print(e)
        sys.exit(-1)



import unittest
from xaal.lib import messages
from xaal.lib import Message,MessageError,MessageType,MessageFactory
from xaal.lib import Device
from xaal.lib import tools,cbor
from xaal.lib.exceptions import MessageParserError  


def fake_message():
    msg = Message()
    msg.msg_type = MessageType.REQUEST
    msg.source = tools.get_random_uuid()
    msg.dev_type= "test.basic"
    msg.action = 'test'
    msg.body = {"foo":"bar"}
    msg.timestamp = messages.build_timestamp()
    return msg

def fake_factory():
    # create the factory
    cypher = tools.pass2key('test')
    mf = MessageFactory(cypher)
    return mf


class TestMessage(unittest.TestCase):

    def test_targets(self):
        msg = Message()
        addr = tools.get_random_uuid()
        # target are stored
        msg.targets = [addr]
        self.assertEqual(msg.targets, [addr,])
        # target is list
        with self.assertRaises(MessageError):
            msg.targets = addr
        # target contains invalid uuid
        with self.assertRaises(MessageError):
            msg.targets = ["foo"]

    def test_display(self):
        msg = fake_message()
        print()
        # dump
        msg.dump()
        print(msg)

    def test_targets_as_string(self):
        msg = fake_message()
        addr = tools.get_random_uuid()
        msg.targets = [addr]
        self.assertEqual(msg.targets_as_string(), [str(addr),] )

    def test_is_request(self):
        msg = fake_message()
        msg.msg_type = 1
        self.assertTrue(msg.is_request())
        msg.msg_type = 2
        self.assertFalse(msg.is_request())


    def test_is_reply(self):
        msg = fake_message()
        msg.msg_type = 2
        self.assertTrue(msg.is_reply())
        msg.msg_type = 1
        self.assertFalse(msg.is_reply()) 

    def test_is_notify(self):
        msg = fake_message()
        msg.msg_type = 0
        self.assertTrue(msg.is_notify())
        msg.msg_type = 1
        self.assertFalse(msg.is_notify())

    def test_is_alive(self):
        msg = fake_message()
        msg.msg_type = 0
        msg.action = 'alive'
        self.assertTrue(msg.is_alive())
        msg.action = 'foo'
        self.assertFalse(msg.is_alive())

    def test_is_request_isalive(self):
        msg = fake_message()
        msg.msg_type = 1
        msg.action = 'is_alive'
        self.assertTrue(msg.is_request_isalive())
        msg.action = 'foo'
        self.assertFalse(msg.is_request_isalive())

    def test_is_attributes_change(self):
        msg = fake_message()
        msg.msg_type = 0
        msg.action = 'attributes_change'
        self.assertTrue(msg.is_attributes_change())
        msg.action = 'foo'
        self.assertFalse(msg.is_attributes_change())

    def test_is_get_attribute_reply(self):
        msg = fake_message()
        msg.msg_type = 2
        msg.action = 'get_attributes'
        self.assertTrue(msg.is_get_attribute_reply())
        msg.action = 'foo'
        self.assertFalse(msg.is_get_attribute_reply())
    
    def test_is_get_description_reply(self):
        msg = fake_message()
        msg.msg_type = 2
        msg.action = 'get_description'
        self.assertTrue(msg.is_get_description_reply())
        msg.action = 'foo'
        self.assertFalse(msg.is_get_description_reply())


class TestMessageFactory(unittest.TestCase):

    def test_encode_decode(self):
        msg1 = fake_message()

        mf = fake_factory()
        data = mf.encode_msg(msg1)
        msg2 = mf.decode_msg(data)

        self.assertEqual(msg1.targets, msg2.targets)
        self.assertEqual(msg1.timestamp, msg2.timestamp)
        self.assertEqual(msg1.source, msg2.source)
        self.assertEqual(msg1.dev_type, msg2.dev_type)
        #FIXME decoded msg_type should be a MessageType
        self.assertEqual(msg1.msg_type.value, msg2.msg_type)
        self.assertEqual(msg1.action, msg2.action)
        self.assertEqual(msg1.body, msg2.body)

    def test_build_alive_for(self):
        dev = Device("test.basic",tools.get_random_uuid())
        mf  = fake_factory()
        data = mf.build_alive_for(dev)
        msg  = mf.decode_msg(data)
        self.assertTrue(msg.is_alive())

    def test_build_error_msg(self):
        dev = Device("test.basic",tools.get_random_uuid())
        mf  = fake_factory()
        data = mf.build_error_msg(dev, 500,"Internal Error")
        msg  = mf.decode_msg(data)
        self.assertTrue(msg.is_notify())
        self.assertEqual(msg.action, 'error')
        self.assertEqual(msg.body['code'], 500)

    def test_decode_cbor_parse_error(self):
        mf = fake_factory()
        with self.assertRaises(MessageParserError):
            mf.decode_msg("not_cbor")

    def test_decode_cbor_fields(self):
        # test missing fields
        mf = fake_factory()
        l = [7,12346.5,[]]
        data=cbor.dumps(l)
        # IndexError
        with self.assertRaises(MessageParserError):
            mf.decode_msg(data)

    def test_decode_filter_func(self):
        msg1 = fake_message()
        filter = lambda msg: False
        mf = fake_factory()
        data = mf.encode_msg(msg1)
        msg2 = mf.decode_msg(data,filter_func=filter)
        self.assertEqual(msg2, None)

    def test_replay_error(self):
        mf = fake_factory()
        # too young
        msg = fake_message()
        target = (msg.timestamp[0] + 60*5, msg.timestamp[1])
        msg.timestamp = target
        data = mf.encode_msg(msg)
        with self.assertRaises(MessageParserError):
            mf.decode_msg(data)
        # too old
        msg.timestamp = messages.build_timestamp()
        target = (msg.timestamp[0] - 60*5, msg.timestamp[1])
        msg.timestamp = target
        data = mf.encode_msg(msg)
        with self.assertRaises(MessageParserError):
            mf.decode_msg(data)

    def test_decode_sanity_check(self):
        mf = fake_factory()
        msg = fake_message()
        # WARNING: Only testing dev_type because alterring source is impossible
        # without rewriting a big chunk of encoding here
        msg.dev_type = "test"
        data = mf.encode_msg(msg)
        with self.assertRaises(MessageParserError):
            mf.decode_msg(data)

    def test_decode_nopayload(self):
        mf = fake_factory()
        msg = fake_message()
        data = mf.encode_msg(msg)
        l=cbor.loads(data)
        l.pop()
        data = cbor.dumps(l)
        with self.assertRaises(MessageParserError):
            mf.decode_msg(data)

    def test_decode_decrypt_error(self):
        mf = fake_factory()
        msg = fake_message()
        data = mf.encode_msg(msg)
        mf.cipher_key = tools.pass2key('bar')
        with self.assertRaises(MessageParserError):
            mf.decode_msg(data)

if __name__ == '__main__':
    unittest.main()



import datetime
import logging
import pprint
import struct
import typing
from enum import Enum
from typing import Any, Optional

import pysodium
from tabulate import tabulate

from . import cbor, tools
from .bindings import UUID
from .config import config
from .exceptions import MessageError, MessageParserError

if typing.TYPE_CHECKING:
    from .devices import Device


logger = logging.getLogger(__name__)

ALIVE_ADDR = UUID("00000000-0000-0000-0000-000000000000")


class MessageType(Enum):
    NOTIFY = 0
    REQUEST = 1
    REPLY = 2


class MessageAction(Enum):
    ALIVE = 'alive'
    IS_ALIVE = 'is_alive'
    ATTRIBUTES_CHANGE = 'attributes_change'
    GET_ATTRIBUTES = 'get_attributes'
    GET_DESCRIPTION = 'get_description'


class Message(object):
    """Message object used for incomming & outgoint message"""

    __slots__ = ['version', 'timestamp', 'source', 'dev_type', 'msg_type', 'action', 'body', '__targets']

    def __init__(self):
        self.version = config.STACK_VERSION  # message API version
        self.__targets = []  # target property
        self.timestamp: tuple = ()  # message timestamp
        self.source: Optional[UUID] = None  # message source
        self.dev_type: Optional[str] = None  # message dev_type
        self.msg_type: Optional[MessageType] = None  # message type
        self.action: Optional[str] = None  # message action
        self.body = {}  # message body

    @property
    def targets(self) -> list:
        return self.__targets

    @targets.setter
    def targets(self, values: list):
        if not isinstance(values, list):
            raise MessageError("Expected a list for targetsList, got %s" % (type(values),))
        for uid in values:
            if not tools.is_valid_address(uid):
                raise MessageError("Bad target addr: %s" % uid)
        self.__targets = values

    def targets_as_string(self) -> list:
        return [str(k) for k in self.targets]

    def dump(self):
        r = []
        r.append(['version', self.version])
        r.append(['targets', str(self.targets)])
        r.append(['timestamp', str(self.timestamp)])
        r.append(['source', self.source])
        r.append(['dev_type', self.dev_type])
        r.append(['msg_type', MessageType(self.msg_type)])
        r.append(['action', self.action])
        if self.body:
            tmp = ""
            for k, v in self.body.items():
                k = k + ":"
                v = pprint.pformat(v, width=55)
                tmp = tmp + "- %-12s %s\n" % (k, v)
            # tmp = tmp.strip()
            r.append(['body', tmp])
        print(tabulate(r, headers=['Field', 'Value'], tablefmt='psql'))

    def __repr__(self) -> str:
        return f"<xaal.Message {id(self):x} {self.source} {self.dev_type} {self.msg_type} {self.action}>"

    def is_request(self) -> bool:
        if MessageType(self.msg_type) == MessageType.REQUEST:
            return True
        return False

    def is_reply(self) -> bool:
        if MessageType(self.msg_type) == MessageType.REPLY:
            return True
        return False

    def is_notify(self) -> bool:
        if MessageType(self.msg_type) == MessageType.NOTIFY:
            return True
        return False

    def is_alive(self) -> bool:
        if self.is_notify() and self.action == MessageAction.ALIVE.value:
            return True
        return False

    def is_request_isalive(self) -> bool:
        if self.is_request() and self.action == MessageAction.IS_ALIVE.value:
            return True
        return False

    def is_attributes_change(self) -> bool:
        if self.is_notify() and self.action == MessageAction.ATTRIBUTES_CHANGE.value:
            return True
        return False

    def is_get_attribute_reply(self) -> bool:
        if self.is_reply() and self.action == MessageAction.GET_ATTRIBUTES.value:
            return True
        return False

    def is_get_description_reply(self) -> bool:
        if self.is_reply() and self.action == MessageAction.GET_DESCRIPTION.value:
            return True
        return False


class MessageFactory(object):
    """Message Factory:
    - Build xAAL message
    - Apply security layer, Ciphering/De-Ciphering chacha20 poly1305
    - Serialize/Deserialize data in CBOR"""

    def __init__(self, cipher_key):
        # key encode / decode message built from passphrase
        self.cipher_key = cipher_key

    def encode_msg(self, msg: Message) -> bytes:
        """Apply security layer and return encode MSG in CBOR
        :param msg: xAAL msg instance
        :type msg: Message
        :return: return an xAAL msg ciphered and serialized in CBOR
        :rtype: CBOR
        """

        # Format security layer
        result = []
        result.append(msg.version)
        result.append(msg.timestamp[0])
        result.append(msg.timestamp[1])
        # list of UUID in bytes format (no Tag here)
        result.append(cbor.dumps([t.bytes for t in msg.targets]))

        # Format payload & ciphering
        buf = []
        if not msg.source:
            raise MessageError("No source address in message")
        if not msg.msg_type:
            raise MessageError("No msg_type in message")
        buf.append(msg.source.bytes)
        buf.append(msg.dev_type)
        buf.append(msg.msg_type.value)
        buf.append(msg.action)
        if msg.body:
            buf.append(msg.body)
        clear = cbor.dumps(buf)
        # Additionnal Data == cbor serialization of the targets array
        ad = result[3]
        nonce = build_nonce(msg.timestamp)
        payload = pysodium.crypto_aead_chacha20poly1305_ietf_encrypt(clear, ad, nonce, self.cipher_key)

        # Final CBOR serialization
        result.append(payload)
        pkt = cbor.dumps(result)
        return pkt

    def decode_msg(self, data: bytes, filter_func: Any = None) -> Optional[Message]:
        """Decode incoming CBOR data and De-Ciphering
        :param data: data received from the multicast bus
        :type data: cbor
        :filter_func: function to filter incoming messages
        :type filter_func: function
        :return: xAAL msg
        :rtype: Message
        """
        # Decode cbor incoming data
        try:
            data_rx = cbor.loads(data)
        except Exception:
            raise MessageParserError("Unable to parse CBOR data")

        # Instanciate Message, parse the security layer
        msg = Message()
        try:
            msg.version = data_rx[0]
            msg_time = data_rx[1]
            targets = cbor.loads(data_rx[3])
            msg.targets = [UUID(bytes=t) for t in targets]
            msg.timestamp = (data_rx[1], data_rx[2])
        except IndexError:
            raise MessageParserError("Bad Message, wrong fields")

        # filter some messages
        if filter_func is not None:
            if not filter_func(msg):
                return None  # filter out the message

        # Replay attack, window fixed to CIPHER_WINDOW in seconds
        now = build_timestamp()[0]  # test done only on seconds ...
        if msg_time < (now - config.cipher_window):
            raise MessageParserError("Potential replay attack, message too old: %d sec" % round(now - msg_time))

        if msg_time > (now + config.cipher_window):
            raise MessageParserError("Potential replay attack, message too young: %d sec" % round(now - msg_time))

        # Payload De-Ciphering
        try:
            ciph = data_rx[4]
        except IndexError:
            raise MessageParserError("Bad Message, no payload found!")

        # chacha20 deciphering
        ad = data_rx[3]
        nonce = build_nonce(msg.timestamp)
        try:
            clear = pysodium.crypto_aead_chacha20poly1305_ietf_decrypt(ciph, ad, nonce, self.cipher_key)
        except Exception:
            raise MessageParserError("Unable to decrypt msg")

        # Decode application layer (payload)
        try:
            payload = cbor.loads(clear)
        except Exception:
            raise MessageParserError("Unable to parse CBOR data in payload after decrypt")
        try:
            msg.source = UUID(bytes=payload[0])
            msg.dev_type = payload[1]
            msg.msg_type = payload[2]
            msg.action = payload[3]
        except IndexError:
            raise MessageParserError("Unable to parse payload headers")
        if len(payload) == 5:
            msg.body = payload[4]

        # Sanity check incomming message
        if not tools.is_valid_address(msg.source):
            raise MessageParserError("Wrong message source [%s]" % msg.source)
        if not tools.is_valid_dev_type(msg.dev_type):
            raise MessageParserError("Wrong message dev_type [%s]" % msg.dev_type)
        return msg

    #####################################################
    # MSG builder
    #####################################################
    def build_msg(
        self,
        dev: Optional['Device'] = None,
        targets: list = [],
        msg_type: Optional[MessageType] = None,
        action: Optional[str] = None,
        body: Optional[dict] = None,
    ):
        """the build method takes in parameters :
        -A device
        -The list of targets of the message
        -The type of the message
        -The action of the message
        -A body if it's necessary (None if not)
        it will return a message encoded in CBOR and ciphered.
        """
        message = Message()
        if dev:
            message.source = dev.address
            message.dev_type = dev.dev_type

        message.targets = targets
        message.timestamp = build_timestamp()

        if msg_type:
            message.msg_type = msg_type
        if action:
            message.action = action
        if body is not None and body != {}:
            message.body = body

        data = self.encode_msg(message)
        return data

    def build_alive_for(self, dev: 'Device', timeout: int = 0) -> bytes:
        """Build Alive message for a given device
        timeout = 0 is the minimum value
        """
        body = {}
        body['timeout'] = timeout
        message = self.build_msg(
            dev=dev, targets=[], msg_type=MessageType.NOTIFY, action=MessageAction.ALIVE.value, body=body
        )
        return message

    def build_error_msg(self, dev: 'Device', errcode: int, description: Optional[str] = None):
        """Build a Error message"""
        message = Message()
        body = {}
        body['code'] = errcode
        if description:
            body['description'] = description
        message = self.build_msg(dev, [], MessageType.NOTIFY, 'error', body)
        return message


def build_nonce(data: tuple) -> bytes:
    """Big-Endian, time in seconds and time in microseconds"""
    nonce = struct.pack(">QL", data[0], data[1])
    return nonce


def build_timestamp() -> tuple:
    """Return array [seconds since epoch, microseconds since last seconds] Time = UTC+0000"""
    utc = datetime.timezone.utc
    epoch = datetime.datetime.fromtimestamp(0, utc)
    timestamp = datetime.datetime.now(utc) - epoch
    return (int(timestamp.total_seconds()), int(timestamp.microseconds))


## This stuff below is for Py2/Py3 compatibility. In the current state of xAAL, we only use
# Py3. This code is here for archive purpose and could be removed in the future.

# for better performance, I choose to use this trick to fix the change in size for Py3.
# only test once.
# if sys.version_info.major == 2:
#    _packtimestamp = lambda t1, t2: (long(t1), int(t2)) # pyright: ignore
# else:
#    _packtimestamp = lambda t1, t2: (int(t1), int(t2))



import asyncio
import logging
import socket
import struct

logger = logging.getLogger(__name__)


class AsyncNetworkConnector(object):
    def __init__(self, addr: str, port: int, hops: int, bind_addr="0.0.0.0"):
        self.addr = addr
        self.port = port
        self.hops = hops
        self.bind_addr = bind_addr
        self._rx_queue = asyncio.Queue()

    async def connect(self):
        loop = asyncio.get_running_loop()
        on_con_lost = loop.create_future()
        self.transport, self.protocol = await loop.create_datagram_endpoint(
            lambda: XAALServerProtocol(on_con_lost, self.receive), sock=self.new_sock()
        )
        # In some conditions (containers), transport is connected but IGMP is delayed (up to 10ms)
        # so we need to wait for IGMP to be really sent.
        await asyncio.sleep(0.05)

    def new_sock(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        try:
            # Linux + MacOS + BSD
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except Exception:
            # Windows
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.bind_addr, self.port))
        mreq = struct.pack("=4s4s", socket.inet_aton(self.addr), socket.inet_aton(self.bind_addr))
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 10)
        sock.setblocking(False)
        return sock

    def send(self, data: bytes):
        self.protocol.datagram_send(data, self.addr, self.port)

    def receive(self, data: bytes):
        self._rx_queue.put_nowait(data)

    async def get_data(self) -> bytes:
        return await self._rx_queue.get()


class XAALServerProtocol(asyncio.Protocol):
    def __init__(self, on_con_lost, on_dtg_recv):
        self.on_con_lost = on_con_lost
        self.on_dtg_recv = on_dtg_recv

    def connection_made(self, transport):
        logger.info("xAAL network connected")
        self.transport = transport

    def error_received(self, exc):
        print("Error received:", exc)
        logger.warning(f"Error received: {exc}")

    def connection_lost(self, exc):
        logger.info(f"Connexion closed: {exc}")
        self.on_con_lost.set_result(True)

    def datagram_send(self, data, ip, port):
        self.transport.sendto(data, (ip, port))

    def datagram_received(self, data, addr):
        # print(f"pkt from {addr}")
        self.on_dtg_recv(data)



# devices.py
class DeviceError(Exception):
    pass

# core.py
class EngineError(Exception):
    pass

class XAALError(Exception):
    pass

class CallbackError(Exception):
    def __init__(self, code, desc):
        self.code = code
        self.description = desc

# messages.py
class MessageParserError(Exception):
    pass

class MessageError(Exception):
    pass

# binding.py
class UUIDError(Exception):
    pass

__all__ = [
    'DeviceError',
    'EngineError',
    'XAALError',
    'CallbackError',
    'MessageParserError',
    'MessageError',
    'UUIDError',
]



import time
import unittest
from xaal.lib import Engine, Device, MessageType, CallbackError
from xaal.lib import engine, tools
from xaal.lib.messages import Message

TEST_PORT = 6666


def new_engine():
    engine = Engine(port=TEST_PORT)
    return engine

class TestEngine(unittest.TestCase):

    def test_devices(self):
        dev = Device("test.basic", tools.get_random_uuid())
        eng = Engine(port=TEST_PORT)
        eng.add_device(dev)
        self.assertEqual(eng.devices,[dev,])
        eng.remove_device(dev)
        self.assertEqual(eng.devices,[])
        eng.add_devices([dev,])
        self.assertEqual(eng.devices,[dev,])

    def test_start_stop(self):
        eng = Engine(port=TEST_PORT)
        dev = Device("test.basic", tools.get_random_uuid())
        eng.add_device(dev)
        eng.start()
        eng.start()  # second start
        self.assertEqual(engine.EngineState.started, eng.state)
        eng.loop()
        eng.stop()
        self.assertEqual(engine.EngineState.halted, eng.state)

    def test_timer(self):
        eng = Engine(port=TEST_PORT)
        t0 = time.time()
        def _exit():
            eng.shutdown()
        eng.add_timer(_exit, 1, 1)
        eng.run()
        t = time.time() - t0
        self.assertTrue(t > 1)
        self.assertTrue(t < 2)

    def test_timer_error(self):
        eng = Engine(port=TEST_PORT)
        eng.start()
        def _error():
            raise CallbackError(500,"test error")
        eng.add_timer(_error,0,1)
        eng.loop()  
        eng.stop()
        self.assertEqual(engine.EngineState.halted, eng.state)

    def test_run_action(self):
        target = Device("test.basic", tools.get_random_uuid())

        def action_1():
            return {"value":"action_1"}

        def action_2(_value=None):
            return {"value":"action_%s" % _value}

        def action_3():
            raise Exception

        target.add_method("action_1", action_1)
        target.add_method("action_2", action_2)
        target.add_method("action_3", action_3)

        msg = Message()
        msg.msg_type = MessageType.REQUEST
        msg.targets = [target.address]
        # simple test method
        msg.action = "action_1"
        result = engine.run_action(msg, target)
        self.assertEqual(result, {"value":"action_1"})
        # test with value
        msg.action = "action_2"
        msg.body = {"value": "2"}
        result = engine.run_action(msg, target)
        self.assertEqual(result, {"value":"action_2"})
        # Exception in method
        msg.action = "action_3"
        with self.assertRaises(engine.XAALError):
            result = engine.run_action(msg, target)
        # unknown method
        msg.action = "missing"
        with self.assertRaises(engine.XAALError):
            result = engine.run_action(msg, target)


if __name__ == "__main__":
    unittest.main()



import uuid

from .exceptions import UUIDError


class UUID:
    def __init__(self, *args, **kwargs):
        self.__uuid = uuid.UUID(*args, **kwargs)

    @staticmethod
    def random_base(digit=2) -> 'UUID':
        """zeros the last digits of a random uuid, usefull w/ you want to forge some addresses
        two digit is great.
        """
        if (digit > 0) and (digit < 13):
            tmp = str(uuid.uuid1())
            st = "%s%s" % (tmp[:-digit], '0' * digit)
            return UUID(st)
        else:
            raise UUIDError

    @staticmethod
    def random() -> 'UUID':
        tmp = uuid.uuid1().int
        return UUID(int=tmp)

    def __add__(self, value: int) -> 'UUID':
        tmp = self.__uuid.int + value
        return UUID(int=tmp)

    def __sub__(self, value: int) -> 'UUID':
        tmp = self.__uuid.int - value
        return UUID(int=tmp)

    def __eq__(self, value) -> bool:
        return self.__uuid == value

    def __lt__(self, value) -> bool:
        return self.__uuid.int < value

    def __gt__(self, value) -> bool:
        return self.__uuid.int > value

    def __str__(self) -> str:
        return str(self.__uuid)

    def __repr__(self) -> str:  # pragma: no cover
        return f"UUID('{self.__uuid}')"

    def __hash__(self) -> int:
        return self.__uuid.__hash__()

    def get(self) -> uuid.UUID:
        return self.__uuid

    def set(self, value: uuid.UUID):
        self.__uuid = value

    @property
    def str(self) -> str:
        return str(self)

    @property
    def bytes(self) -> bytes:
        return self.__uuid.bytes


class URL:
    def __init__(self, value):
        self.__url = value

    def __eq__(self, value):
        return self.__url == value

    def __str__(self):
        return str(self.__url)

    def __repr__(self) -> str:  # pragma: no cover
        return f"URL('{self.__url}')"

    def set(self, value: str):
        self.__url = value

    def get(self) -> str:
        return self.__url

    @property
    def str(self):
        return str(self)

    @property
    def bytes(self):
        return self.__url


classes = [UUID, URL]



import unittest
from xaal.lib import tools
from xaal.lib import Device,DeviceError,Attribute,Engine
from xaal.lib.devices import Attributes


def new_dev() -> Device:
    addr = tools.get_random_uuid()
    dev = Device("light.basic",addr=addr)
    return dev

class TestAttribute(unittest.TestCase):

    def test_default(self):
        attr = Attribute("foo",dev=new_dev(),default=1)
        self.assertEqual(attr.value,1)
        attr.value = 12
        self.assertEqual(attr.value,12)

    def test_with_engine(self):
        dev = new_dev()
        dev.engine = Engine()
        attr = Attribute("foo",dev=dev,default=1,)
        attr.value = 12
        self.assertEqual(attr.value,12)

class TestAttributes(unittest.TestCase):

    def test_default(self):
        dev = new_dev()
        attr = Attribute("foo",dev=dev,default=1)
        attrs = Attributes()
        attrs.append(attr)
        self.assertEqual(attrs[0].value,1)
        self.assertEqual(attrs["foo"],1)

        with self.assertRaises(KeyError):
            attrs["bar"]

        attrs[0] = Attribute("bar",dev=dev,default=2)
        self.assertEqual(attrs[0].value,2)

        attrs["bar"] = 12
        self.assertEqual(attrs["bar"],12)

        with self.assertRaises(KeyError):
            attrs["nop"]=12


class TestDevice(unittest.TestCase):

    def test_init(self):
        dev_type="light.basic"
        addr = tools.get_random_uuid()
        dev = Device(dev_type=dev_type,addr=addr)
        self.assertEqual(dev.dev_type, dev_type)
        self.assertEqual(dev.address, addr)

    def test_address(self):
        dev = new_dev()
        # address
        addr = tools.get_random_uuid()
        dev.address = addr
        self.assertEqual(dev.address, addr)
        # none address
        dev.address = None
        self.assertEqual(dev.address, None)
        # invalid address
        with self.assertRaises(DeviceError):
            dev.address = "foo"

    def test_dev_type(self):
        dev = new_dev()
        dev.dev_type = "foo.basic"
        self.assertEqual(dev.dev_type, "foo.basic")
        with self.assertRaises(DeviceError):
            dev.dev_type = "foo"

    def test_version(self):
        dev = new_dev()
        # version 
        dev.version = 12
        self.assertEqual(dev.version, '12')
        dev.version = None
        self.assertEqual(dev.version, None)

    def test_url(self):
        dev = new_dev()
        dev.url = "http://foo.bar"
        self.assertEqual(dev.url, "http://foo.bar")
        dev.url = None
        self.assertEqual(dev.url, None)

    def test_attributes(self):
        dev = new_dev()
        attr = Attribute("foo",dev=dev,default=1)
        # add
        dev.add_attribute(attr)
        self.assertEqual(len(dev.attributes),1)
        self.assertEqual(dev.attributes[0],attr)
        # del
        dev.del_attribute(attr)
        self.assertEqual(len(dev.attributes),0)
        # new
        dev.new_attribute("foo",default=2)
        self.assertEqual(len(dev.attributes),1)
        self.assertEqual(dev.attributes[0].value,2)
        # get
        self.assertEqual(dev.get_attribute("foo").value,2)
        self.assertEqual(dev.get_attribute("bar"),None)
        
        # set
        attr = Attribute("bar",dev=dev,default=3)
        dev.attributes =Attributes([attr,])
        self.assertEqual(len(dev.attributes),1)
        self.assertEqual(dev.attributes[0].value,3)
        # only accepts Attributes not list 
        with self.assertRaises(DeviceError):
            dev.attributes = [attr,]

    def test_get_attributes(self):
        dev = new_dev()
        dev.new_attribute("foo",default=1)
        # _get_attributes
        self.assertEqual(dev._get_attributes()["foo"],1)
        data = dev._get_attributes(["foo","bar"])
        self.assertEqual(len(data),1)
        self.assertEqual(data["foo"],1)

    def test_get_description(self):
        dev = new_dev()

        dev.vendor_id = 0x1234
        dev.product_id = 0x1234
        dev.version = '2f'
        dev.url = "http://foo.bar"
        dev.schema = "http://schemas.foo.bar/schema.json"
        dev.info = "FooBar"
        dev.hw_id = 0xf12
        group = tools.get_random_uuid()
        dev.group_id = group

        dev.unsupported_methods = ["foo_func"]
        dev.unsupported_attributes = ["foo_attr"]
        dev.unsupported_notifications = ["foo_notif"]

        data = dev._get_description()
        self.assertEqual(data["vendor_id"], 0x1234)
        self.assertEqual(data["product_id"], 0x1234)
        self.assertEqual(data["version"], '2f')
        self.assertEqual(data["url"], "http://foo.bar")
        self.assertEqual(data["schema"], "http://schemas.foo.bar/schema.json")
        self.assertEqual(data["info"], "FooBar")
        self.assertEqual(data["hw_id"],0xf12)
        self.assertEqual(data["group_id"],group)

        self.assertEqual(data["unsupported_methods"],["foo_func"])
        self.assertEqual(data["unsupported_notifications"],["foo_notif"])
        self.assertEqual(data["unsupported_attributes"],["foo_attr"])

    def test_methods(self):
        dev = new_dev()
        # device has two methods by default
        self.assertEqual(len(dev.methods),2)
        self.assertEqual(len(dev.get_methods()),2)
        def func():pass
        dev.add_method("foo",func)
        self.assertEqual(dev.methods["foo"],func)

    def test_dump(self):
        dev = new_dev()
        dev.info = 'FooBar'
        dev.new_attribute("foo",default=1)
        dev.dump()

    def test_alive(self):
        import time
        now = time.time()
        dev = new_dev()
        dev.alive_period = 2 
        dev.update_alive()
        self.assertTrue(dev.next_alive > now)
        self.assertTrue(dev.next_alive < now+3)
        self.assertEqual(dev.get_timeout(),4)

    def test_send_notification(self):
        # not really a test here, just to check if the method is called
        # check engine unittest for futher tests
        dev = new_dev()
        dev.engine = Engine()
        dev.send_notification("foo")
        

if __name__ == '__main__':
    unittest.main()


import unittest
from xaal.lib import bindings


class TestUUID(unittest.TestCase):


    def test_add_sub(self):
        uuid = bindings.UUID("12345678-1234-1234-1234-123456789012")
        uuid += 1
        self.assertEqual(uuid,bindings.UUID("12345678-1234-1234-1234-123456789013"))
        uuid -= 1
        self.assertEqual(uuid,bindings.UUID("12345678-1234-1234-1234-123456789012"))

    def test_random(self):
        uuid = bindings.UUID.random()
        self.assertIsInstance(uuid,bindings.UUID)

        uuid = bindings.UUID.random_base()
        self.assertIsInstance(uuid,bindings.UUID)

        with self.assertRaises(bindings.UUIDError):
            uuid = bindings.UUID.random_base(0)

    def test_get_set(self):
        uuid1 = bindings.UUID("12345678-1234-1234-1234-123456789012")
        data = uuid1.get()
        uuid2 = bindings.UUID.random()
        uuid2.set(data)
        self.assertEqual(uuid1,uuid2)

    def test_str(self):
        uuid = bindings.UUID("12345678-1234-1234-1234-123456789012")
        self.assertEqual(str(uuid),"12345678-1234-1234-1234-123456789012")
        self.assertEqual(uuid.str,"12345678-1234-1234-1234-123456789012")

    def test_bytes(self):
        uuid = bindings.UUID("12345678-1234-1234-1234-123456789012")
        self.assertEqual(uuid.bytes,b'\x124Vx\x124\x124\x124\x124Vx\x90\x12')


class TestURL(unittest.TestCase):

    def test_url(self):
        url = bindings.URL("http://bar.com")
        self.assertEqual(str(url), "http://bar.com")
        url.set("http://foo.com")
        print(url)
        self.assertEqual(str(url), "http://foo.com")
        self.assertEqual(url.str, "http://foo.com")
        self.assertEqual(url.get(), "http://foo.com")
        self.assertEqual(url.bytes, "http://foo.com")


if __name__ == '__main__':
    unittest.main()


import logging
import select
import socket
import struct
import time
from enum import Enum
from typing import Optional

logger = logging.getLogger(__name__)


class NetworkState(Enum):
    disconnected = 0
    connected = 1


class NetworkConnector(object):
    UDP_MAX_SIZE = 65507

    def __init__(self, addr: str, port: int, hops: int, bind_addr="0.0.0.0"):
        self.addr = addr
        self.port = port
        self.hops = hops
        self.bind_addr = bind_addr
        self.state = NetworkState.disconnected

    def connect(self):
        try:
            self.__connect()
        except Exception as e:
            self.network_error(e)

    def __connect(self):
        logger.info("Connecting to %s:%s" % (self.addr, self.port))
        self.__sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        try:
            # Linux + MacOS + BSD
            self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except Exception:
            # Windows
            self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.__sock.bind((self.bind_addr, self.port))
        mreq = struct.pack("=4s4s", socket.inet_aton(self.addr), socket.inet_aton(self.bind_addr))
        self.__sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        self.__sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, self.hops)
        self.state = NetworkState.connected

    def disconnect(self):
        logger.info("Disconnecting from the bus")
        self.state = NetworkState.disconnected
        self.__sock.close()

    def is_connected(self):
        return self.state == NetworkState.connected

    def receive(self) -> bytes:
        packt = self.__sock.recv(self.UDP_MAX_SIZE)
        return packt

    def __get_data(self) -> Optional[bytes]:
        r = select.select([self.__sock,], [], [], 0.02)
        if r[0]:
            return self.receive()
        return None

    def get_data(self) -> Optional[bytes]:
        if not self.is_connected():
            self.connect()
        try:
            return self.__get_data()
        except Exception as e:
            self.network_error(e)

    def send(self, data: bytes):
        if not self.is_connected():
            self.connect()
        try:
            self.__sock.sendto(data, (self.addr, self.port))
        except Exception as e:
            self.network_error(e)

    def network_error(self, ex: Exception):
        self.disconnect()
        logger.info("Network error, reconnect..%s" % ex.__str__())
        time.sleep(5)
